#!/usr/bin/python
################################################################################
#
# File:         bibexport.py
# RCS:          $Header: /u/testbed/CVSROOT/dldev/src/interbib2007/Scripts/bibexport.py,v 1.15 2007/08/04 03:40:05 paepcke Exp $
# Description:  Classes that export bibliographies
# Author:       Andreas Paepcke
# Created:      Sat Sep 23 09:12:10 1995
# Modified:     Fri Aug 03 17:06:03 2007 (Andreas Paepcke) paepcke@SNUGGLES
# Language:     Python
# Package:      N/A
# Status:       Experimental (Do Not Distribute)
#
# (c) Copyright 2007, Andreas Paepcke, all rights reserved.
#
################################################################################
#
# Revisions:
#  In BibTeX exporter: used to pry the entry type from an entry's class name.
# Thu Aug 02 11:06:31 2007 (Andreas Paepcke) paepcke@EMIL
#  Added EndnoteExporter.
# Sun Jul 29 14:45:45 2007 (Andreas Paepcke) paepcke@EMIL
#  Changed calls to method convert() into convertAttr()
#  Changed that to use the entries' recordType field.
#   Old: #entryType = entry.__class__ .__name__
#   New: entryType = entry.recordType()
# Sun Jul 29 13:10:41 2007 (Andreas Paepcke) paepcke@EMIL
#  
################################################################################

import string
import bibdb
import bibconvert
import sys

#########################################################################
#
# Class BibliographyExporter 
#
# Description: this class is an abstract class.  It defines the interface
#              to a Bibliography Exporter and implements common 
#              functionality
#
# Usage:       This class is intended to be inhereted by a child 
#              The child of this class 
#              should define a method with the signature:
#                emitTranslation(self, entry, outFd, errFd)
#              This methods writes in outFd the "entry", if there are 
#              any errors/warnings they should be written to errFd. The
#              only method that it's guarantee to be in outFd and errFd 
#              is "write"
#
# Public Methods:
#  __init__(filename, verbosity)
#  exportBibliography( outFd, errFd )
#
#########################################################################

class BibliographyExporter:

  db = 0
  warnings = []
  errors = []

  def __init__(self, fileName='', db=0, verbosity='verbose'):
    if fileName != '':
      try:
	self.db = bibdb.BibDB(dbStore = fileName)
      except IOError:
	raise IOError, "Error opening database"
    else:
      self.db = db
    self.verbosity = verbosity

  # end __init__

  def exportBibliography( self, outFd, errFd ):

    try:

      self.warnings = []
      keys = self.db.keys()
      for key in keys:
	entry = self.db[key]
	self.emitTranslation( entry, outFd, errFd )

    finally:

      # Do error and warning reporting according to
      # the verbosity setting:

      if self.verbosity == 'verbose':
	for warning in self.warnings:
	  errFd.write(warning)
	  errFd.write('\n')

  def emitLinkInfo(self, linkValue, outFd):
    for i in range(len(linkValue)):
      link = linkValue[i]
      outFd.write(link.ascii())
      if i < len(linkValue) - 1:
	outFd.write(',')

  # end exportBibliography



#########################################################################
#
# Class BibtexExporter
#
# Description: this class implements an exporter to the Bibtex format
#
# Public Methods:
#  __init__(filename, verbosity)
#  exportBibliography( outFd, errFd ) (inhereted from BibliographyExporter)
#
#########################################################################


class BibtexExporter(BibliographyExporter):

  def __init__(self, fileName='', db=0, verbosity='verbose'):
    BibliographyExporter.__init__(self, 
				  fileName=fileName, 
				  db = db,
				  verbosity=verbosity)
    self.db.setOutputCharSet(bibconvert.LatexCharSet)
  # end __init__

  def emitTranslation(self, entry, outFd, errFd):

    attributeNames = entry.keys()

    # Get entry's record type:
    entryType = entry[bibconvert.typeFieldName]

    outFd.write('@' + entryType + '{')

    outFd.write(entry[bibconvert.keyFieldName] + '\n')

    # Authors are separated by '|' in the entry:

    if "author" in attributeNames:
      authors = string.splitfields(entry["author"],"|")
      outFd.write("  ,author = {" + authors[0])
      for author in authors[1:]:
	outFd.write(" and " +  author)
      outFd.write("}\n")

    if "title" in attributeNames:
      outFd.write("  ,title = {"+entry["title"]+'}\n')

    # The special Interbib <anchor>:<environment>:<dest> facility:

    if bibconvert.linksFieldName in attributeNames:

      outFd.write("  ," + bibconvert.linksFieldName + " = ")
      outFd.write('"')
      self.emitLinkInfo(entry[bibconvert.linksFieldName], outFd)
      outFd.write('"\n')

    # We output the 'url' field (if present) as: "\url{...}"
    # to get the LaTeX typesetting nice:

    if bibconvert.urlFieldName in attributeNames:
      outFd.write("  ," + bibconvert.urlFieldName + " = ")
      outFd.write('"')
      outFd.write("\\url{")
      outFd.write(entry[bibconvert.urlFieldName] + '}"\n')

    for attributeName in attributeNames:

      # Skip handling the special fields we did manually above,
      # or that just shouldn't be translated:
      if attributeName == "author":
	pass
      elif attributeName == bibconvert.keyFieldName:
	pass
      elif attributeName == "title":
	pass
      elif attributeName == bibconvert.linksFieldName:
	pass
      elif attributeName == bibconvert.urlFieldName:
	pass
      elif attributeName == bibconvert.typeFieldName:
	pass
      elif entry[attributeName] == " " or entry[attributeName] == "" :
	pass
      else:
	outFd.write("  ," + attributeName + " = {"+entry[attributeName] + '}\n')
    outFd.write('}\n\n')
  # end emitTranslation

#########################################################################
#
# Class ReferExporter
#
# Description: this class implements an exporter to the Refer format
#
# Public Methods:
#  __init__(filename, verbosity)
#  exportBibliography( outFd, errFd ) (inhereted from BibliographyExporter)
#
#########################################################################

class ReferExporter(BibliographyExporter):

  def __init__(self, fileName='', db=0, verbosity='verbose'):
    BibliographyExporter.__init__(self, 
				  fileName=fileName, 
				  db = db,
				  verbosity=verbosity)
    self.db.setOutputCharSet(bibconvert.Latin1CharSet)
    self.attrConverter = bibconvert.ReferAttributeConverter()
    self.recTypeConverter = bibconvert.RecordTypeNameConverterEndnoteInterbib()

  # end __init__

  def emitTranslation(self, entry, outFd, errFd):

    attributeNames = entry.keys()

    # Write out the entry type ("Journal Article", etc.):

    referRecordType = self.recTypeConverter.getEndnoteRecTypeFromInterBib (entry[bibconvert.typeFieldName])
    outFd.write('%0  ' + referRecordType  + '\n')

    if "author" in attributeNames:
      authors = string.splitfields(entry["author"],'|')
      for author in authors:
	outFd.write('%A  ' + author.strip() + '\n')        # Clean up surrounding white space

    if "title" in attributeNames:
      outFd.write("%T  " + entry["title"].strip() + '\n')  # Clean up surrounding white space

    if bibconvert.linksFieldName in attributeNames:
      outFd.write('%1  ')
      self.emitLinkInfo(entry[bibconvert.linksFieldName], outFd)
      outFd.write('\n')

    # Go through the attr names and translate them,
    # skipping over the fields we took care of manually above:

    for attributeName in attributeNames:
      if attributeName == "author":
	pass
      elif entry[attributeName] == " " or entry[attributeName] == "" :
	pass
      elif attributeName == "title":
	pass
      elif attributeName == "month":
	pass
      elif attributeName == "day":
	pass
      elif attributeName == bibconvert.linksFieldName:
	pass
      elif attributeName == bibconvert.typeFieldName:
	pass
      elif attributeName == "year":
	if "month" in attributeNames:
	  month = entry["month"] + " "
	else:
	  month = ""
	if "day" in attributeNames:
	  day = entry["day"] + " "
	else:
	  day = ""
	date = day + month + entry["year"]
	outFd.write('%D  '+date+'\n')
      else:
	try:
	  outFd.write('%' + self.attrConverter.reverseConvertAttr(attributeName))
	  outFd.write("  " + entry[attributeName]+'\n')
	except bibconvert.NonTranslatableAttribute, attr:
	  self.warnings.append('Warning: in ' +\
			       entry[bibconvert.keyFieldName] +\
			       ", " +\
			       attr.message() +\
			       ' = "' +\
			       entry[attr.message()] +\
			       '" was not translated')
    outFd.write('\n')

  # end emitTranslation

#########################################################################
#
# Class EndnoteExporter
#
# Description: this class implements an exporter to the Endnote format
#
# Public Methods:
#  __init__(filename, verbosity)
#  exportBibliography( outFd, errFd ) (inhereted from BibliographyExporter)
#
#########################################################################

class EndnoteExporter(ReferExporter):

  def __init__(self, fileName='', db=0, verbosity='verbose'):
    ReferExporter.__init__(self, 
			   fileName=fileName, 
			   db = db,
			   verbosity=verbosity)
    self.db.setOutputCharSet(bibconvert.Latin1CharSet)
    self.attrConverter = bibconvert.EndnoteAttributeConverter()
    self.recTypeConverter = bibconvert.RecordTypeNameConverterEndnoteInterbib()

  # end __init__

#
# Some tests
#

#a = BibtexExporter( "/u/crespo/temp/my-paper" )
#a.exportBibliography(sys.stdout, sys.stderr)

#print "****"

#a = ReferExporter( "/u/crespo/temp/my-paper" )
#a.exportBibliography(sys.stdout, sys.stderr)

