#! /usr/bin/python
################################################################################
#
# File:         bibframe.py
# RCS:          $Header: /u/testbed/CVSROOT/dldev/src/interbib2007/Scripts/bibframe.py,v 1.1 2007/07/24 20:24:57 paepcke Exp $
# Description:  Transforming MIF Files into MIFs with citations resolved + ref. List
# Author:       Andreas Paepcke, Stanford University
# Created:      Sun Oct 29 22:45:06 1995
# Modified:     Mon Jul 23 17:39:13 2007 (Andreas Paepcke) paepcke@SNUGGLES
# Language:     Python
# Package:      N/A
# Status:       Experimental (Do Not Distribute)
#
# (c) Copyright 1995, Stanford University, all rights reserved.
#
################################################################################
#
# Revisions:
#
# Sun Jul 22 07:52:10 2007 (Andreas Paepcke) paepcke@SNUGGLES
#  Converted to the new exception convention (no string exceptions; new calls for each)
# Tue Jul  2 12:33:09 1996 (Andreas Paepcke) paepcke@Phoebe.Stanford.EDU
#  In processFile: added test for empty outFile: This is needed because the test
#  used to be just for the existence of outFd. But outFd now always is initialized
#  to stdout.
# Sun Jan 14 09:32:03 1996 (Andreas Paepcke) paepcke@AndreasHomePC
#  Added setDocumentCitationDelimiters
# Tue Dec 19 16:51:28 1995 (Andreas Paepcke) paepcke@Stingray
#  Added support for inserting bibliography in a particular place in the MIF doc.
# Mon Dec 18 15:15:45 1995 (Andreas Paepcke) paepcke@Phoebe.Stanford.EDU
#  Corrected the redirect of standard error for -q option.
# Wed Dec  6 22:54:02 1995 (Andreas Paepcke) paepcke@Stingray
#  Added acceptance of refer files.
################################################################################

# Usage: bibframe [-sqrci]               See 'usage' string below for details
#                 <inFileName>
#                 <outFileName>
#                 <bibFile1> <bibFile2>...

# Takes inFile which must be a Framemaker or a .mif file. Uses the bibFiles
# to resolve bibliographic refs in inFile. Writes to outFile another MIF or
# Framemeaker file which has citations resolved and has a reference list
# inserted into it. The citations in the infile must be of the form
# '[smith89, brad90]', where the abbrevs are keys in the bibliographies.
# Currently, BibTex and Refer is supported. They are recognized by a .bib
# or .ref extension.

# You can use the separate shell-level program 'bibload' to turn bib files
# into native InterBib databases (recognized by filename with no
# extension). If you use a Bibtex or Refer file often, it's worth
# converting them into several, or one all-comprising InterBib database,
# because then your bibliography generation will be significantly faster.

# The -i option allows you to specify a word that occurs in your document.
# InterBib will insert the bibliography after the last occurrence of that
# word (search is case insensitive). If you do not specify -i or -s (see
# below), the bibliography will be stuck at the end. This will often not
# work properly, and the bibliography will not show up in your output. So
# using one of the options is highly recommended.

# The -s option causes the bibliography to be put into a separate file (the
# infile name plus '-refs<n>'). The user must then cut/paste or import the
# reference list manually from Framemaker (using the filemenu-->import file
# in Framemaker). If the -s option is not used, the user is warned to check
# whether the list made it into the output.

# If when you import, the reference list is in a different size font,
# Don't panic. Just highlight the whole list and reapply the 'Body' par
# format to it. (just pull down the 'Body' button in the button bar at the
# top and re-select 'body' from the pull-down menu)

# The -q option causes all output to be suppressed. We return 1 if all went
# well, we return 0 if errors occurred.

# The style of citations and the style of references in the ref list default
# to 'NumericCiteAppearanceOrder' and 'GenericStyle' respectively. The former
# makes citations of the form '[1, 2, 3]'. The latter produces references
# of that conform to Claire Van Leunen's 'Handbook for Scholars'.

# These settings can be changed with the -c and -r options respectively.
# (Alternatively, setting the env vars '$BIBCITEFORMAT' and
# '$BIBREFFORMAT' will change them as well. Legal citation formats
# are:

#
#   - NumericCiteAppearanceOrder
#   - NumericCiteAuthorAlphaOrder
#   - AlphaCiteAppearanceOrder
#   - AlphaCiteKeyAlphaOrder
#   - AlphaCiteAuthorAlphaOrder
#

# Legal reference formats are currently 'GenericStyle' and 'IEEEStyle'.

# NOTE: the bib importers are very observant. So you may see a lot of
#       warnings and errors. Don't worry about them, unless the program
#       then fails to resolve references. In that case, check the errors
#       to see whether the corresponding entry was not imported.
#

# Performance examples: For a 25 page document with 9 half-page images,
#                       worst-case (frm->mif->frm, numericCite/AuthorAlphaOrder
#                       which requires an additional pass, and insertion of
#                       reference list after a specified word in the doc. Bu
#                       bibliography precompiled as InterBib db): 2min22sec
#
#                       Same doc, but only involving MIF files: 1min53sec

import sys
import bibutils
import bibdb
import bibdocxform
import bibconvert
import getopt
import os
import string
import tempfile

#####################################################
#
# Constants
#
#####################################################

citationEnvVar = '$BIBCITEFORMAT'
refEnvVar = '$BIBREFFORMAT'
legalCitationFormats = ['NumericCiteAppearanceOrder',
			'NumericCiteAuthorAlphaOrder',
			'AlphaCiteAppearanceOrder',
			'AlphaCiteKeyAlphaOrder',
			'AlphaCiteAuthorAlphaOrder']
legalRefFormats =      ['GenericStyle',
			'IEEEStyle']

#####################################################
#
# Exceptions
#
#####################################################

class MissingConverterFacility(bibutils.InterbibException):
  pass


#####################################################
#
# Class Definitions
#
#####################################################


class bibMaker:

  def __init__(self,
	       targetEnv = 'mif',
	       outFd = sys.stdout,
	       errorFd = sys.stderr,
	       runQuiet = 0):
    self.inFile = ''
    self.outFile = ''
    self.outFd = outFd
    sys.stdout = errorFd
    self.originalInFile = ''
    self.originalOutFile = ''
    self.refListFile = ''
    self.originalRefListFile = ''
    self.bibFiles = []
    self.targetEnv = targetEnv
    self.errors = []
    self.warnings = []
    self.citationFormat = None
    self.refFormat = None
    self.citationOpener = '['
    self.citationCloser = ']'
    self.dbs = []
    self.runQuiet = runQuiet
    self.filesToDelete = []
    self.insertionMarker = ''
    self.xformer = None
    self.usage = "Usage: bibframe [-q][-s][-c NumericCiteAppearanceOrder |\n\
                            NumericCiteAuthorAlphaOrder |\n\
                            AlphaCiteAppearanceOrder |\n\
                            AlphaCiteKeyAlphaOrder |\n\
                            AlphaCiteAuthorAlphaOrder]\n\
			[-r GenericStyle |\n\
			    IEEEStyle]\n\
			[-i <Word in doc after which to insert the reference list>]\n\
                <inFileName> <outFileName>\n\
                <bibFile1> <bibFile2>..." 

    # Initialize the temp file name generation:
    tempfile.template = 'interbib-'

  def inputErrorCheck(self):

    try:
      optlist, args = getopt.getopt(sys.argv[1:], 'qsc:r:i:')
    except getopt.error, info:
      if not self.runQuiet:
	bibutils.report("Illegal option: " + str(info))
	bibutils.report(self.usage)
      return(0)
    for option in optlist:
      if option[0] == '-s':
	# Just remember that we are to keep references separate:
        self.refListFile = 'foo'
      if option[0] == '-q':
	self.runQuiet = 1
      if option[0] == '-c':
	self.citationFormat = option[1]
      if option[0] == '-r':
	self.refFormat = option[1]
      if option[0] == '-i':
	self.insertionMarker = option[1]

    if len(args) < 3:
      if not self.runQuiet:
        bibutils.report("Not enough args\n" + self.usage)
      return(0)

    self.inFile = args[0]
    self.outFile = args[1]

    if self.inFile == self.outFile:
      if not self.runQuiet:
        bibutils.report("Input file must be different from output file.")
      return(0)
    if not os.path.isfile(self.inFile):
      if not self.runQuiet:
        bibutils.report("File '" + self.inFile + "' not found.")
      return(0)

    # Make sure we have a full path for infile and outfile:
    if os.path.basename(self.inFile) == self.inFile:
      # Full path not given:
      self.inFile = os.getcwd() + '/' + self.inFile
    else:
      # Make sure all env vars and ~ are resolved:
      self.inFile = os.path.expanduser(os.path.expandvars(self.inFile))

    if os.path.basename(self.outFile) == self.outFile:
      # Full path not given:
      self.outFile = os.getcwd() + '/' + self.outFile
    else:
      # Make sure all env vars and ~ are resolved:
      self.outFile = os.path.expanduser(os.path.expandvars(self.outFile))

    # Compute name of reflist file if -s option was given. It's the
    # full-path input file name, with '-refs.mif' appended:

    if self.refListFile:
      refFileName = os.path.splitext(self.inFile)[0] + '-refs.mif'
      if os.path.isfile(refFileName):
	uniquifier = 0
        while os.path.isfile(refFileName):
  	  refFileName = os.path.splitext(self.inFile)[0] +\
			'-refs' +\
			str(uniquifier) +\
			'.mif'
	  uniquifier = uniquifier + 1
      self.refListFile = refFileName

    self.bibFiles = args[2:]
    return(1)

  def defaultSetup(self):

    if not self.citationFormat:
      self.citationFormat = os.path.expandvars(citationEnvVar)
      if self.citationFormat == citationEnvVar:
        # There was no env var for citation format:
        self.citationFormat = 'NumericCiteAppearanceOrder'
    if self.citationFormat not in legalCitationFormats:
      if not self.runQuiet:
	bibutils.report("Bad citation format spec:", str(self.citationFormat))
	bibutils.report(self.usage)
      return(0)
  
    if not self.refFormat:
      self.refFormat = os.path.expandvars(refEnvVar)
      if self.refFormat == refEnvVar:
        # There was no env var for reference format:
        self.refFormat = 'GenericStyle'
    if self.refFormat not in legalRefFormats:
      if not self.runQuiet:
	bibutils.report("Bad reference format spec:", str(self.refFormat))
	bibutils.report(self.usage)
      return(0)

    return(self.fixInFileFormats())

  def fixInFileFormats(self):

    try:
      if isRtfFile(self.inFile):
	self.xformer = bibdocxform.RTFDocTransformer()
	return(1)
      # If infile is not a mif file, it's either Frame or HTML. In that case:
      # convert to MIF:
      if isMifFile(self.inFile):
	self.xformer = bibdocxform.MIFDocTransformer()
      elif isFrameFile(self.inFile):
	tmpName = tempfile.mktemp() + '.mif'
	if not self.runQuiet:
	  bibutils.report("Converting input file from frame native to MIF...")
	if not frame2mif(self.inFile, outFile=tmpName, runQuiet=self.runQuiet):
	  if not self.runQuiet:
	    bibutils.report("Unable to create temporary MIF file for: " + self.inFile)
	  return(0)
	else:
	  if not self.runQuiet:
	    bibutils.report("Done.")
	  self.filesToDelete.append(tmpName)
	    
	self.originalInFile = self.inFile
	self.inFile = tmpName
        self.xformer = bibdocxform.MIFDocTransformer()

      else:
	self.xformer = bibdocxform.HTMLDocTransformer()
	return(1)
  
      if self.outFile and not isMifFileName(self.outFile):
	tmpName = tempfile.mktemp() + '.mif'
	self.originalOutFile = self.outFile
	self.outFile = tmpName
	self.filesToDelete.append(tmpName)
  
      if self.refListFile and not isMifFileName(self.refListFile):
	tmpName = tempfile.mktemp() + '.mif'
	self.originalRefListFile = self.refListFile
	self.refListFile = tmpName
	self.filesToDelete.append(tmpName)

    except MissingConverterFacility, info:
      if not self.runQuiet:
	bibutils.report("Problem with file conversion: " + info.message())
      return(0)
    
    return(1)
    

  # This converts the output file(s) to the user's intended final
  # format. It also cleans up temp files:

  def fixOutFileFormats(self):

    if self.originalOutFile:
      if isFrameFileName(self.originalOutFile):
	if not self.runQuiet:
	  bibutils.report("Converting output file from MIF to frame native...")
	if not mif2frame(self.outFile,
			 outFile=self.originalOutFile,
			 runQuiet=self.runQuiet):
	  if not self.runQuiet:
	    bibutils.report("Unable to create final frame file: " + self.originalOutFile)
	  return(0)
	else:
	  self.outFile = self.originalOutFile

    if self.originalRefListFile:
      if isFrameFileName(self.originalRefListFile):
	if not self.runQuiet:
	  bibutils.report("Converting reference list file from MIF to frame native...")
	if not mif2frame(self.refListFile,
			 outFile=self.originalRefListFile,
			 runQuiet=self.runQuiet):
	  if not self.runQuiet:
	    bibutils.report("Unable to create final frame file for reference list: " + self.originalRefListFile)
	  return(0)
	else:
	  self.refListFile = self.originalRefListFile

    if not self.runQuiet:
      if self.refListFile:
        bibutils.report("NOTE: You must now import " + self.refListFile + " into " + self.outFile + " to get the final document!")
      else:
        bibutils.report("NOTE: Check whether the reference list is in the doc. If not,\n    redo using the -s option")
    return(1)


  def bibliographyImport(self,
			 duplicatePolicy = bibdb.autoResolveRecordConflicts,
			 dbIdentifiers = []):

    for i in range(len(self.bibFiles)):
      bibFile = self.bibFiles[i]
      if len(dbIdentifiers) > i:
	dbId = dbIdentifiers[i]
      else:
	dbId = ''
      if not os.path.isfile(bibFile):
        if not self.runQuiet:
	  bibutils.report("File '" + bibFile + "' not found.")
	return(0)
      # Check whether bib file is supported:
      bibsys = checkBibFile(bibFile)
      if not bibsys:
	if not self.runQuiet:
	  bibutils.report("Only BibTex, Refer and InterBib dbs supported right now: Bib files must have '.bib' or no extension")
	return(0)
      if bibsys == 'interbib':
	# an already imported InterBib citation db:
	self.dbs.append(bibdb.BibDB(dbStore = bibFile, permissions = 'r'))
	self.dbs[-1].name = dbId
        if not self.runQuiet:
          bibutils.report('\nImported bibliography ' + dbId)
      else:
	self.dbs.append(bibdb.BibDB())
	self.dbs[-1].name = dbId
        if not self.runQuiet:
          bibutils.report('\nBeginning to import bibliography ' + dbId)
	self.dbs[-1].importBibliography(sourceFile = bibFile,
					bibsystem = bibsys,
 				        duplicatePolicy = duplicatePolicy,
					verbosity = not self.runQuiet)
        if not self.runQuiet:
          bibutils.report('done importing bibliography ' + dbId)
	self.errors = self.errors + self.dbs[-1].errors
	self.warnings = self.warnings + self.dbs[-1].warnings

    return(1)

  # fileHeader is an optional fileHeader to be placed before the
  # final file:
  def processFile(self, fileHeader='', annote=0):

     self.xformer.setDocumentCitationDelimiters(self.citationOpener,
					        self.citationCloser)
     
     if isRtfFile(self.inFile):
       for db in self.dbs:
	 db.setOutputCharSet(bibconvert.RTFCharSet)

     if not self.runQuiet:
       bibutils.report('Starting preparation of your document...')
     if self.outFd and not self.outFile:
       res = self.xformer.transformDocument(inFileName = self.inFile,
				       outFd = self.outFd,
				       refListFileName = self.refListFile,
				       citationStyle = self.citationFormat,
				       refStyle = self.refFormat,
				       insertionPt = self.insertionMarker,
				       dbs = self.dbs,
				       annote = annote,
				       systemFileHeader = fileHeader)
     else:
       res = self.xformer.transformDocument(inFileName = self.inFile,
				       outFileName = self.outFile,
				       refListFileName = self.refListFile,
				       citationStyle = self.citationFormat,
				       refStyle = self.refFormat,
				       insertionPt = self.insertionMarker,
				       dbs = self.dbs,
				       systemFileHeader = fileHeader)
    
     if not self.runQuiet:
       bibutils.report('Done preparing your document.')
     self.errors = self.errors + self.xformer.errors
     self.warnings = self.warnings + self.xformer.warnings
     return(res)
    
  def setDocumentCitationDelimiters(self, citationOpener, citationCloser):
    self.citationOpener = citationOpener
    self.citationCloser = citationCloser

  def cleanUp(self):
    for file in self.filesToDelete:
      if os.path.isfile(file):
	os.system('rm ' + file)

# Return a string with the bibliography format of the file, as derived
# from the file name extension: If no extension, assume file is an InterBib
# db. If it's a '.bib' extension, assume BibTeX. Else, return 0:

def checkBibFile(bibFile):
  components = os.path.splitext(bibFile)
  if components[1] == '.bib':
    return('bibtex')
  elif components[1] == '.ref':
    return('refer')
  # an InterBib database?:
  elif components[1] == '':
    return('interbib')
  else:
    return(0)

def frame2mif(inFile, outFile='', runQuiet=0):
  if not isInPath('fmbatch'):
    if os.path.isfile('./fmbatch'):
      fmbatchPath = './'
    else:
      raise MissingConverterFacility, "Framemaker utility 'fmbatch' not available at server.\n\
      This utility is required to convert Framemaker format to MIF. Submit MIF if you can."
  else:
    fmbatchPath = ''
  if not outFile:
    # make outFile be inFile with the extension replaced by '.mif':
    outFileDir = os.path.split(inFile)[0]
    outFileName = os.path.basename(inFile)
    outFileBase = os.path.splitext(outFileName)[0]
    outFile = outFileDir + '/' + outFileBase + '.mif'
  else:
    outFileDir = os.path.split(outFile)[0]
  if runQuiet:
    outputRedirect = ' >/dev/null 2>/dev/null'
  else:
    outputRedirect = ''
  # The following solution used to work on HP-UX, but doesn't
  # on Solaris, at least not w/ Frame 5.5. I leave it here for reference.
  #if os.system('echo "Open ' +\
	#       inFile +\
	#       '\\nSaveAs m ' +\
	#       inFile +\
	#       ' ' +\
	#       outFile +\
	#       '\\nQuit" | ' +\
	#       fmbatchPath + 'fmbatch' +\
	#       outputRedirect) == 0 and\
	#       os.path.isfile(outFile):
  # Create a file called 'fmbatchFile' in the tmp dir below us.
  # That file is then passed to fmbatch for execution:
  os.system('echo "Open ' + inFile +
	    '\nSaveAs m ' + inFile + ' ' + outFile +
	    '\nQuit\n" > tmp/fmbatchFile')
  if os.system(fmbatchPath + 'fmbatch tmp/fmbatchFile ' +\
	       outputRedirect) == 0 and\
	       os.path.isfile(outFile):

    return(1)
  else:
    return(0)
    
def mif2frame(inFile, outFile='', runQuiet=0):
  if not isInPath('fmbatch'):
    raise MissingConverterFacility, "Framemaker utility 'fmbatch' not available at server.\n\
    This utility is required to convert MIF format to Framemaker. Request MIF if you can."
  if not outFile:
    # make outFile be inFile with the extension replaced by '.frm':
    outFileDir = os.path.split(inFile)[0]
    outFileName = os.path.basename(inFile)
    outFileBase = os.path.splitext(outFileName)[0]
    outFile = outFileDir + '/' + outFileBase + '.frm'
  if runQuiet:
    outputRedirect = ' >/dev/null 2>/dev/null'
  else:
    outputRedirect = ''
  if os.system('echo "Open ' +
	       inFile +
	       '\\nSaveAs d ' +
	       inFile +
	       ' ' +
	       outFile +
	       '\\nQuit" | fmbatch' +
	       outputRedirect) == 0 and\
	       os.path.isfile(outFile):
    return(1)
  else:
    return(0)

def isMifFileName(fileName):
  return(os.path.splitext(os.path.basename(fileName))[1] == '.mif')

def isFrameFileName(fileName):
  return(os.path.splitext(os.path.basename(fileName))[1] == '.frm')

def isRtfFileName(fileName):
  return(os.path.splitext(os.path.basename(fileName))[1] == '.rtf')

# We assume that the file exists!
def isMifFile(fileName):
  fd = open(fileName)
  ln = fd.readline()
  fd.close()
  if ln[:8] == '<MIFFile':
    return(1)
  else:
    return(0)
  

# We assume that the file exists!
def isFrameFile(fileName):
  fd = open(fileName)
  ln = fd.readline()
  fd.close()
  if ln[:10] == '<MakerFile':
    return(1)
  else:
    return(0)
  #return(os.path.splitext(os.path.basename(fileName))[1] == '.frm')

def isRtfFile(fileName):
  fd = open(fileName)
  ln = fd.readline()
  fd.close()
  if string.lower(ln[:5]) == '{\\rtf':
    return(1)
  else:
    return(0)
  #return(os.path.splitext(os.path.basename(fileName))[1] == '.rtf')
  
def isInPath(fileName):
  for path in string.split(os.environ['PATH'], ':'):
    if os.path.isfile(path + '/' + fileName):
      return(1)
  return(0)

def main():
  myBibMaker = bibMaker()
  if not myBibMaker.inputErrorCheck():
    myBibMaker.cleanUp()
    return(0)
  if not myBibMaker.defaultSetup():
    myBibMaker.cleanUp()
    return(0)
  if not myBibMaker.bibliographyImport():
    myBibMaker.cleanUp()
    return(0)
  if not myBibMaker.processFile():
    myBibMaker.cleanUp()
    return(0)
  if not myBibMaker.fixOutFileFormats():
    myBibMaker.cleanUp()
    return(0)
  myBibMaker.cleanUp()
  return(1)
  
# If we are called from a shell, run now. If we are being imported,
# do nothing:
#if sys.argv[0]:
#  main()

