################################################################################
#
# File:         bibrefgen.py
# RCS:          $Header: /u/testbed/CVSROOT/dldev/src/interbib2007/Scripts/bibrefgen.py,v 1.3 2007/07/31 03:39:50 paepcke Exp $
# Description:  Generating reference lists from BibService databases
# Author:       Andreas Paepcke, Stanford University
# Created:      Thu Sep 28 15:01:31 1995
# Modified:     Mon Jul 30 20:38:49 2007 (Andreas Paepcke) paepcke@SNUGGLES
# Language:     Python
# Package:      N/A
# Status:       Experimental (Do Not Distribute)
#
# (c) Copyright 1995, Stanford University, all rights reserved.
#
################################################################################
#
# Revisions:
#
# Sun Jul 22 07:56:47 2007 (Andreas Paepcke) paepcke@SNUGGLES
#  Converted to the new exception convention (no string exceptions; new calls for each)
# Fri Jul 20 11:37:44 2007 (Andreas Paepcke) paepcke@Shark.Stanford.EDU
#  Migrated from deprecated regex to re.
# Tue Apr  1 11:10:50 1997 (Andreas Paepcke) paepcke@Haddock.Stanford.EDU
#  Fixed bug that caused blowup when trying to add to tagged strings which
#  had links in their tags.
# Sat Aug 10 17:07:03 1996 (Andreas Paepcke) paepcke@Walrus
#  In generateRef: exception info used to be just 'styleName'. When a
#  non-string was passed in, this bombed. So added str conversion.
# Thu Feb 22 16:27:24 1996 (Andreas Paepcke) paepcke@AndreasHomePC
#  Fixed bug causing duplicate particle in refs if it ended with an
#  abbreviation (like 'D.C.' AND the next particle was a period.
# Thu Feb 22 06:35:31 1996 (Andreas Paepcke) paepcke@AndreasHomePC
#  Added FullAuthorStyle.
# Thu Feb 22 06:34:45 1996 (Andreas Paepcke) paepcke@AndreasHomePC
#  fixed Generic and IEEE conferenceStyle classes to include 'Generic' and
#  'IEEE' in the class names.
# Sat Feb  3 22:19:10 1996 (Andreas Paepcke) paepcke@AndreasHomePC
#  Added option to include/exclude annotations in references.
# Mon Jan  1 12:31:34 1996 (Andreas Paepcke) paepcke@AndreasHomePC
#  Made error and warning reports use bibutils.report()
# Mon Dec 25 21:07:07 1995 (Andreas Paepcke) paepcke@Stingray
#  Made missing biblio fields into a warning, rather than an error.
################################################################################

# Public Classes and Methods:
#
#   ReferenceGenerator
#     attr: errors
#     attr: warnings
#     generateRef
#     showErrors
#     showWarnings
#   TaggedStr
#     attr: str
#     attr: tag
#     find
#     rfind
#     index
#      ... (other string functions)


# To do:
#
#  - Update comment at head of file

import string
import types
import copy
import re
import bibutils
import bibimport
import bibconvert

#####################################################
#
# Global, Public Constants
#
#####################################################

supportedStyles = ['GenericStyle', 'IEEEStyle', 'FullAuthorStyle']

# String tags. If you change the type of their value, change method __setattr__ of
# class TaggedStr as well!

plain   = 'plain'
italic  = 'italic'
bold    = 'bold'
link    = 'link'

mandatory      = 1
advised        = 2
optional       = 0

#####################################################
#
# Exceptions
#
#####################################################

class IllegalEntryType(bibutils.InterbibException):
  pass

class MissingField(bibutils.InterbibException):
  pass

class UnsupportedStyleType(bibutils.InterbibException):
  pass

class BadTagValue(bibutils.InterbibException):
  pass

class BadReferenceParticle(bibutils.InterbibException):
  pass

class BadParameter(bibutils.InterbibException):
  pass

class IncompatibleTags(bibutils.InterbibException):
  pass

#####################################################
#
# Private Constants
#
#####################################################

twoChrAbbrevMatch = re.compile(' [A-Z].[A-Z].')
twoChrStateMatch = re.compile(' [A-Z][A-Z].')

#####################################################
#
# Classes
#
#####################################################


# ********THESE COMMENTS NEEDS UPDATE!!!!!!! Things Much Simpler Now*********

# Explain particles, style classes and subclasses, synthesizeReference method,
# spec constructor methods (calling them after override of particles!! Explain
# smartSeparators and supersmartSeparators. Explain to inherit from the most
# specific first. Example:
#    ReferenenceGen->IEEEStyle->bookIEEEStyle(bookGenericStyle, IEEEStyle)
#
# Particular reference styles are all child classes of ReferenceGenerator. 
# This parent provides primitives such as author name manipulation, error
# handling convenciences, etc. You can also inherit as well from someone
# else (ex: inbookIEEEStyle inherits from inbookGenericStyle). But in
# your __init__ method: must call most specific parent init method last
# to get that init method's overrides in there!

# Each ref style class is called something like 'IEEEStyle'. There is one
# pubmedium-specific subclass for each kind of reference. Ex:
# 'articleIEEEStyle', 'bookIEEEStyle', etc. Each reference is organized
# into four major parts: Author info, Title info, Publication info and
# Notes Info.  Each of these gets handled by each of the pubmedium-specific
# subclasses, though sometimes subclassing can be used if there are no
# differences. So: articleIEEEStyle and bookIEEEStyle each handle all four
# parts of a reference (using the methods described below). If a style
# FooStyle uses the same reference format for inproceedings and for inbook,
# it can have inbookFooStyle inherit from inproceedingsFooStyle.

# The methods handling the reference parts are:
#   1. genAuthorInfo(entry)
#   2. genTitleInfo(entry)
#   3. genPubInfo(entry)
#   4. genNotesInfo(entry)

# Each of these methods returns an array of TaggedStr objects. The first
# and last of these tagged strings must assume that they have no separator
# at the start and end, respectively. The separating period or comma is
# put in by a method joinParts(authorInfo, titleInfo, pubInfo, notesInfo).
# **********REVISE TO FIT NEW MECHANISM************
# So, pubInfo might look like this:
#   [(plain/'vol: '), (bold/'3, '), (plain/'p. 10, 1979')]
# Note that the volume has no leading comma separating it from the title
# that will be ahead of it. And the year does not have a closing period.
# Those are added by joinParts. But *within* the pubInfo in this example,
# the delimiters are there. The genPubInfo method is repsonsible for not
# putting multiple commas when some info is missing, etc.

# To add a new style FooStyle, create a class of that name, inheriting from
# ReferenceGenerator. Create subclasses articleFooStyle, bookFooStyle, etc.
# for each of the bib entry types. They all inherit from FooStyle.
# For each of these subclassses, decide whether you want to have them behave
# like one of their siblings, in which case you make them inherit from that
# sibling, instead of from FooStyle directly. Otherwise, implement the
# five methods above for the subclass.

#######################################################

######################################
#
# BibReference
#
######################################

# This class holds a seq of tagged strings that together make up the
# target application-independent bib reference:

class BibReference:
  refSpec = []
  charSet = bibconvert.FrameCharSet

  def __init__(self, refSpec=[], charSet=bibconvert.FrameCharSet):
    self.refSpec = refSpec
    self.charSet = charSet
  
  def prependTstr(self, tstr):
    self.refSpec = [tstr] + self.refSpec

  def appendTstr(self, tstr):
    self.refSpec.append(tstr)


######################################
#
# RefParticle
#
######################################

# RefParticle objects are Used to specify one piece of reference. For
# example the piece in the publication details: ', vol. <vol-number>'. The
# separator is the separation from any particle that may have preceded this
# one.  Any non-tag field (except for the fieldName field) can either be a
# string or a function which is called at runtime with the field value and
# the entry being worked on. The function must then return a string.
# Example for the 'pages' part of the publication detail: The leader
# attribute in a particle for the pages field could be set to
# self.pageOrPages. This method would be called at runtime with the value
# of the pages field of the reference entry. It could check whether it's a
# single page or a page range and return either 'p. ' or 'pp. ' as
# appropriate.
#
# Example: leader = self.pageOrPages
#
# ****Explain the case of a function in fldName ******
#
# Also, the field name may either be a single field name, or several fields
# separated by:
#     '_|_': this causes the value of the first non-empty field to be inserted
#     '_&_': this causes the non-empty values of all fields to be concatenated using
#            conjunctSeparator in between.
#    '_&&_': this is like _&_, but concatenation happens only when ALL fields
#            are non-empty. Otherwise no output is generated.
#
# The following is legal: '[self.fixAuthors, author] _|_ editor'
#
# Ref particles can be declared optional (the default), mandatory or advised. If the
# field on a mandatory particle is empty, an exception is raised (quietly if verbose is
# off). If and advised field is empty, a warning is generated but the reference is
# still generated. This is used when the omission of a field may cause anomalies in the
# reference, such as a comma following a period.

class RefParticle:
  separatorTag = ''  # Ex: plain
  separator = ''     # Ex: ', '
  leaderTag = ''     # Ex: bold
  leader = ''        # Ex: '('
  fieldTag = ''      # Ex: bold
  fieldName = ''     # Ex: 'volume'
  trailerTag = ''    # Ex: plain
  trailer = ''       # Ex: ')'. For a journal article's number particle this could be ')'
  existenceReq = 0   # Ex: mandatory Could be voluntary instead. These are constants.
  conjunctSeparator = ''
  fieldValue = ''    # this is filled in at runtime.

  def __init__(self,
	       separatorTag = plain,      
	       separator = '',   
	       leaderTag = plain,
	       leader = '',      
	       fieldTag = plain,
	       fieldName = '',
	       trailerTag = plain,
	       trailer = '',
	       existenceReq = optional,
	       conjunctSeparator = ' '):
    self.separatorTag    = separatorTag
    self.separator       = separator
    self.leaderTag       = leaderTag
    self.leader          = leader
    self.fieldTag        = fieldTag
    self.fieldName       = fieldName
    self.trailerTag      = trailerTag
    self.trailer         = trailer
    self.existenceReq    = existenceReq
    self.conjunctSeparator = conjunctSeparator


######################################
#
# ReferenceGenerator: ABSTRACT CLASS!
#
######################################


class ReferenceGenerator:
  errors = []
  warnings = []
  verbosity = 1
  smartSeparators = 1      # Suppress separators in Ref Particles if prev char was a sep?
  supersmartSeparators = 1 # Handle "This is a title," as well?
  separators = '.,;:?!'     # Chars that are separators
  separatorsForCapitalization = '.:?!' # seps that should cause caps after it.
  lastSeparator = []       # Last (TaggedStr) separator inserted. (Could be a composite)


  def __init__(self, verbosity = 1):

    self.lastSeparator = []
    self.verbosity = verbosity

    self.authorParticle = RefParticle(
      fieldName     = '[self.fixAuthors, author] _|_ editor',
      trailer       = '. ',
      existenceReq  = mandatory)

    self.titleParticle  = RefParticle(
      fieldName     = 'title',
      trailer       = '. ',
      existenceReq  = mandatory)
 
    self.chapterParticle = RefParticle(
      separator     = ', ',
      fieldName     = '[self.handleChapter, chapter]')

    self.pageParticle = RefParticle(
      separator     = ', ',
      fieldName     = '[self.pageOrPages, pages]')
  
    self.numParticle = RefParticle(
      separator     = ', ',
      fieldName     = '[self.handleNumber, number]')

    self.seriesParticle = RefParticle(
      separator     = '. ',
      fieldName     = 'series _&_ [self.handleVolume, volume]',
      conjunctSeparator = ', ')

    self.publisherParticle  = RefParticle(
      separator     = '. ',
      fieldName     = 'publisher _|_ organization _|_ school _|_ institution')

    self.editionParticle = RefParticle(
      separator     = ', ',
      fieldName     = 'edition')

    self.addressParticle  = RefParticle(
      separator     = ', ',
      fieldName     = 'address')

    self.dateParticle = RefParticle(
      separator     = ', ',
      fieldName     = 'month _&_ year',
      conjunctSeparator = ', ')

    self.notesParticle  = RefParticle(
      separator     = '. ',
      fieldName     = 'howpublished _&_ note',
      conjunctSeparator = '. ')
  
    self.annoteParticle  = RefParticle(
      separator     = '. ',
      fieldName     = 'annote _&_ comments _&_ abstract',
      conjunctSeparator = '. ')
  
    self.constructAuthorSpec()
    self.constructTitleSpec()
    self.constructPubInfoSpec()
    self.constructNotesSpec()
    self.constructAnnoteSpec()

  def constructAuthorSpec(self, spec=[]):
    # For future changes to spec format (like making it into an object),
    # we have this no-op call if caller passes array of particles, because
    # right now, array of particles *are* the format of specs:
    if spec:
      self.authorSpec = spec
    else:
      self.authorSpec = [self.authorParticle]
  
  def constructTitleSpec(self, spec=[]):
    # For future changes to spec format (like making it into an object),
    # we have this no-op call if caller passes array of particles, because
    # right now, array of particles *are* the format of specs:
    if spec:
      self.titleSpec = spec
    else:
      self.titleSpec = [self.titleParticle]

  def constructPubInfoSpec(self, spec=[]):
    # For future changes to spec format (like making it into an object),
    # we have this no-op call if caller passes array of particles, because
    # right now, array of particles *are* the format of specs:
    if spec:
      self.pubInfoSpec = spec
    else:
      self.pubInfoSpec = [self.pageParticle,
	 		  self.seriesParticle,
			  self.numParticle,
			  self.publisherParticle,
			  self.editionParticle,
		          self.addressParticle,
		          self.dateParticle]

  def constructNotesSpec(self, spec=[]):
    # For future changes to spec format (like making it into an object),
    # we have this no-op call if caller passes array of particles, because
    # right now, array of particles *are* the format of specs:
    if spec:
      self.notesSpec = spec
    else:
      self.notesSpec = [self.notesParticle]

  def constructAnnoteSpec(self, spec=[]):
    # For future changes to spec format (like making it into an object),
    # we have this no-op call if caller passes array of particles, because
    # right now, array of particles *are* the format of specs:
    if spec:
      self.annoteSpec = spec
    else:
      self.annoteSpec = [self.annoteParticle]

  def showErrors(self):
    if self.errors:
      for error in self.errors:
        bibutils.report(error)
    else:
      bibutils.report("No errors.")

  def showWarnings(self):
    if self.warnings:
      for warn in self.warnings:
        bibutils.report(warn)
    else:
      bibutils.report("No warnings.")

  # This is the top level entry: call it with an entry and a style name. It
  # returns an BibReference instance which contains the bibliographic reference.
  # That object can be passed to one of the compilers:

  def generateRef(self, entry, styleName, annote=0):
    if styleName not in supportedStyles:
      self.handleError("Style '" +\
			 str(styleName) +\
			 "' is not supported.",
		       UnsupportedStyleType, type(entry))
      
    # Get an object of the most specific subclass (named after the style
    # name and the entry type) such as 'articleIEEEStyle', or the parent
    # IEEEStyle:

    className = entry[bibconvert.typeFieldName] + styleName
    if self.classExists(className):
      style = eval(className)()
      style.verbosity = self.verbosity
    else:
      style = ReferenceGenerator()
      style.verbosity = self.verbosity

    # Produce a list of tagged strings and package them into a BibReference
    # object. Since we want to stay target platform-independent, we need
    # to set the entry up to use Frame char set. Then we reset that afterwards:

    oldOutputCharSet = entry.outputCharSet
    entry.outputCharSet = bibconvert.FrameCharSet

    try:
      refList = style.synthesizeReference(entry, annote=annote)
      return(BibReference(refList, entry.nativeCharSet))
    finally:
      # Restore the original output char set:
      entry.outputCharSet = oldOutputCharSet
    
  # This method can be overridden in the subclasses (styles and their entry-specific
  # children). If a style can live just by new authorSpec,titleSpec,pubSpec and
  # notesSpec attributes, great. No methods are needed in the subclasses then:

  def synthesizeReference(self, entry, annote=0):
    if annote:
      return(self.genRefPieceFromSpec(entry, self.authorSpec\
				      + self.titleSpec\
				      + self.pubInfoSpec\
				      + self.notesSpec\
				      + self.annoteSpec))
    else:
      return(self.genRefPieceFromSpec(entry, self.authorSpec\
				      + self.titleSpec\
				      + self.pubInfoSpec\
				      + self.notesSpec))

	   
  # Called automatically from the author spec with the authors.  Put
  # authors into last name last form, separate them by commas, and put the
  # 'and' before the last author. Return an array of TaggedStr:

  def fixAuthors(self, authors, entry, particle):
    if not authors:
      return([])
    tags = [plain] + bibutils.getLinkTags(entry, 'author')
    authors = string.splitfields(authors, bibimport.authorSeparator)
    res = bibutils.getAuthorLastNameLast(authors[0])
    if not string.strip(res):
      # Got author which is just spaces (some people do that to fool bibtext into
      # thinking that there is an author):
      return([])
    if len(authors) > 1:
      lastAuthor = authors[-1]
    else:
      lastAuthor = ""
    for author in authors[1:]:
      if author == lastAuthor:
	# Van Leunen says: if three or more authors, put a comma before the closing 'and':
	if len(authors) > 2:
	  res = res + ', and ' + bibutils.getAuthorLastNameLast(author)
	else:
	  res = res + ' and ' + bibutils.getAuthorLastNameLast(author)
      else:
        res = res + ', ' + bibutils.getAuthorLastNameLast(author)
    return([TaggedStr(tags, res)])

  def handleWarning(self, warnStr):
    self.warnings.append(warnStr)
    if self.verbosity:
      bibutils.report("Warning: " + warnStr)

  def handleError(self, errStr, exception, excArg=""):
    self.errors.append(errStr)
    if self.verbosity:
      raise exception, excArg

  def handleMissingFldError(self, entry, fldName):
    self.warnings.append("Warning: Required field '"\
		         + fldName\
		         + "' not present in '"\
		         + entry[bibconvert.keyFieldName] \
		         + "'. (I'll still do it)")
    if self.verbosity:
      bibutils.report(self.warnings[-1])

    
  # Given an array of RefParticle, generate an array of TaggedStr with the
  # reference:

  def genRefPieceFromSpec(self, entry, spec):
    if not spec:
      return([])
    # Check spec parameter:
    for specElement in spec:
      if type(specElement) <> types.InstanceType:
        self.handleError("Quantity '" +\
			 str(specElement) +\
			 "' is not a reference particle.",
		         BadParameter, specElement)
    res = []
    self.lastSeparator = []
    # Get all the field values. Handle mandatory ones that are empty:
    for particle in spec:

      # Must evaluate separator and leader before the field value, so
      # that functions which compute based on most recent separator
      # get the correct separator:
      
      if type(particle.separator) <> types.StringType:
	self.handleError("Separators must be strings: '" +\
			 str(particle.separator),
			 BadParameter, particle)
			 
      # Get the separator processed, so that functions computing the
      # field value will see it. But since the value might end up
      # empty, we don't add the sep to the official result yet:
      prelimRes = []
      prelimRes = self.addSeparatorIntelligently(res,\
					         [TaggedStr(particle.separatorTag,\
						            particle.separator)])

      particle.fieldValue = self.getFieldValue(particle, entry)
      # If no value and the fld is mandatory, throw up:
      if (particle.existenceReq == mandatory) and not particle.fieldValue:
	self.handleMissingFldError(entry, particle.fieldName)
      if (particle.existenceReq == advised) and not particle.fieldValue:
	self.handleWarning("Should have non-empty '" + \
			   particle.fieldName +\
			   "' in entry '" +\
			   entry[bibconvert.keyFieldName] +\
			   "'.")
      # If the field was non-empty, go for it:
      if particle.fieldValue:
	fldValue = particle.fieldValue
	res = prelimRes

	if type(particle.leader) == types.StringType:	
	  res.append(TaggedStr(particle.leaderTag, particle.leader))
	else:
	  self.checkFunctionAppSpec(particle.leader)
	  res + particle.leader(fldValue, entry, particle)
	res = res + fldValue
	if type(particle.trailer) == types.StringType:	
	  res = self.addSeparatorIntelligently(res,\
					       [TaggedStr(particle.trailerTag,\
						          particle.trailer)])
	else:
	  self.checkFunctionAppSpec(particle.trailer)
          res = self.addSeparatorIntelligently(res,\
					       particle.trailer(fldValue,\
								entry,\
								particle))
    res = self.addSeparatorIntelligently(res, [TaggedStr(plain, '.')])
    return(self.compactTaggedStrings(res))

  # Take an array TaggedStr(s) which comprise a particle separator (already
  # worked out, if there was a function call involved, etc.) And take a
  # (possibly empty) array of TaggedStr (usually from previously processed
  # particles). If self.smartSeparators is turned on, we do the following:
  # If the previous TaggedStr is a member of self.separators (things like
  # comma, period, etc), then we do not add the separator to the array. If
  # the last char is not a separator, we check whether
  # self.supersmartSeparators is on. If yes, we check whether this is a
  # case of "This is a title," which some bib styles call for. We count
  # that comma as a previous separator.
  #
  # We try to be smart about cases where double separators *are* required:
  #
  #     Bluebeard does it Again. Addison-Wesley, N.Y., pp 10-14.
  #
  # But we need to be carefule to do this OK as well:
  #
  #     Bluebeard....  it Again. Addison-Wesley, N.Y. This will blow
  #       your socks of.
  #
  # So we check for two-cap-char endings with period at end and maybe
  # a period in the middle. If the previous tstr ends with an abbrev and
  # the new separator is something other than a period, we pretend the
  # previous period doesn't exist. But if the new sep is a period, we
  # make sure there will be a space after the abbrev, other than that,
  # we call that period that ends the abbrev our last sep:

  def addSeparatorIntelligently(self, prevResults, separatorTstr):
    if (not self.smartSeparators) or (not prevResults):
      # If there is a sep and it's non-empty, remember it as the last one:
      if separatorTstr and separatorTstr[-1]:
	self.lastSeparator = separatorTstr
      return(prevResults + separatorTstr)
    if (not separatorTstr):
      return(prevResults)
    # Get copy of last tstr that is non-empty:
    for i in range(len(prevResults)):
      if prevResults[-i]:
	originalPrevTstr = prevResults[-i]
	prevTstr = originalPrevTstr.strip()
	if prevTstr:
	  lastNonEmptyTstrIndx = -i
	  break
    if not prevTstr:
      # Empty or only spaces in entire previous history:
      if separatorTstr and separatorTstr[-1]:
        self.lastSeparator = separatorTstr
      return(prevResults + separatorTstr)
    prevChr = prevTstr[-1]
    if prevChr in self.separators:
      # Check for prev tstr ending in an abbrev:
      if self.endsWithAbbrev(prevTstr):
	if separatorTstr and separatorTstr[-1].strip().str == '.':
	  # Separator is again a period. Add a space after the previous
	  # tstr if necessacry, and remember the new sep as the last one,
	  # but don't put in the second period:
          self.lastSeparator = separatorTstr
	  if originalPrevTstr[-1] <> ' ':
	    prevResults[lastNonEmptyTstrIndx] = originalPrevTstr + ' '
          return(prevResults)
	else:
          if separatorTstr and separatorTstr[-1]:
            self.lastSeparator = separatorTstr
          return(prevResults + separatorTstr)
      # Don't add the separator. Check whether we have a space after the
      # already-there prev. string. Then return:
      if originalPrevTstr[-1] <> ' ':
	prevResults[lastNonEmptyTstrIndx] = originalPrevTstr + ' '
      return(prevResults)
    # Check whether it's a separator followed by '"':
    if (not self.supersmartSeparators)\
       or (len(prevTstr) < 2)\
       or ((not prevChr == '"')\
	   and (not prevChr == "'")):
      if separatorTstr and separatorTstr[-1]:
        self.lastSeparator = separatorTstr
      return(prevResults + separatorTstr)
    # We know the prev str ended with a quote mark. Check char before that for
    # being a separator:
    secondToLastChr = prevTstr[-2]
    if secondToLastChr in self.separators:
      return(prevResults)
    else:
      if separatorTstr and separatorTstr[-1]:
        self.lastSeparator = separatorTstr
      return(prevResults + separatorTstr)

  # Takes a tstr and returns 1 if it ends in an abbreviation of either the
  # form "D.C." or "CA.". That is two cap letters with a period in between,
  # or two cap letters with a period after. This is a heuristic used to
  # allow double separators in refs after abbrevs. As in:
  #     Bluebeard does it Again. Addison-Wesley, N.Y., pp 10-14

  def endsWithAbbrev(self, tstr):
    if len(tstr) < 4:
      return(0)
    if twoChrStateMatch.match(tstr.str, len(tstr)-4) > -1:
      return(1)
    if len(tstr) < 5:
      return(0)
    else:
      return(twoChrAbbrevMatch.match(tstr.str, len(tstr)-5) > -1)


  # Takes a ref particle and an entry. Gets the value of the field of the particle.
  # That field can be any of the following kinds of examples:
  #    - 'author'
  #    - 'author _|_ corporateAuthor'
  #    - 'notes _&_ comments'
  #    - 'notes _&&_ comments'
  #    - '[self.fixAuthors, author]'
  #    - '[self.fixAuthors, author] _|_ title'
  #
  # See comments with class RefParticle for explanations of _|_, _&_ and _&&_.
  # The notation '[<func>, <fld-name>]' means to call the function <func> with
  # the value of field <fld-name> to get the final result.
  # We return an array of TaggedStr. It will be empty if all TaggedStr's are
  # empty.

  def getFieldValue(self, particle, entry):
    if type(particle) <> types.InstanceType:
      self.handleError("Quantity '" +\
		       str(particle) +\
		       "' is not a reference particle.",
		       BadParameter, particle)
    if string.find(particle.fieldName, '_|_') > -1:
      fldNames = string.splitfields(particle.fieldName, '_|_')
      for fldName in fldNames:
        value = self.getFldOrApplyFunction(entry, fldName, particle.fieldTag, particle)
	if value:
	  return(value)
      return([])
    elif string.find(particle.fieldName, '_&&_') > -1:
      fldNames = string.splitfields(particle.fieldName, '_&&_')
      res = []
      foundAtLeastOne = 0
      conjunct = [TaggedStr(plain, particle.conjunctSeparator)]
      # Remember the old sep in case we set it below but then never insert it.
      oldLastSeparator = self.lastSeparator
      for fldName in fldNames:
	# Set last sep to be the conjunct separator (if non-empty):
	if conjunct[0] and foundAtLeastOne:
	  self.lastSeparator = conjunct
        value = self.getFldOrApplyFunction(entry, fldName, particle.fieldTag, particle)
	if value:
	  if foundAtLeastOne:
	    res = res + conjunct + value
	    # We definitely know that the conjunct is now the latest separator:
	    oldLastSeparator = conjunct
	  else:
	    res = res + value
	    foundAtLeastOne = 1
	else:
	  self.lastSeparator = oldLastSeparator
	  return([])
      return(res)
    elif string.find(particle.fieldName, '_&_') > -1:
      fldNames = string.splitfields(particle.fieldName, '_&_')
      res = []
      foundAtLeastOne = 0
      conjunct = [TaggedStr(plain, particle.conjunctSeparator)]
      # Remember the old sep in case we set it below but then never insert it.
      oldLastSeparator = self.lastSeparator
      for fldName in fldNames:
	# Set last sep to be the conjunct separator (if non-empty). 
	if conjunct[0] and foundAtLeastOne:
	  self.lastSeparator = conjunct
        value = self.getFldOrApplyFunction(entry, fldName, particle.fieldTag, particle)
	if value:
	  if foundAtLeastOne:
	    res = res + conjunct + value
	    # We definitely know that the conjunct is now the latest separator:
	    oldLastSeparator = conjunct
	  else:
	    res = res + value
	    foundAtLeastOne = 1
	else:
	  self.lastSeparator = oldLastSeparator
      return(res)
    else:
      # It's not an expression involving _|_, _&_ or _&&_:
      return(self.getFldOrApplyFunction(entry, \
					particle.fieldName, \
					particle.fieldTag,\
					particle))


  # Takes a bib entry and a particle fld spec (see comments for
  # getFieldValue and RefParticle). Returns a possibly empty array of
  # TaggedStr. There will be no empty TaggedStr in the array.
  # For values it gets right out of fields, it only counts them if
  # they are not all empty:

  def getFldOrApplyFunction(self, entry, fldValueSpec, tag, particle):
    if type(tag) <> types.ListType:
      tag = [tag]
    fldValueSpec = string.strip(fldValueSpec)

    # Is the field value spec has a [ in it, we call a func to get a value:

    if string.find(fldValueSpec, '[') > -1:
      fnFldPair = self.checkFunctionAppSpec(fldValueSpec)
      # have a '[function, fldName]' string. Turn that string into an
      # array and apply the function (which returns an array of TaggedStr):
      value = self.applyFnToFldValue(entry, fnFldPair, particle)
      return(self.compactTaggedStrings(value))
    else:
      value = bibutils.getEntryFldSafely(entry, fldValueSpec)
      # If field value only has blanks, count it as not there.
      if string.strip(value):
	tag = tag + bibutils.getLinkTags(entry, fldValueSpec)
        return([TaggedStr(tag, value)])
      else:
	return([])

  # Check whether the argument passed to it is one of the following:
  #  - a function or method
  #  - a string of the form '[<func>, <str>]'
  # If not, it generates an error. Else it returns the function object in the first
  # case, or an array with a function object and a string in the second case:
  def checkFunctionAppSpec(self, funcAppSpec):
    if type(funcAppSpec) == types.FunctionType or\
       type(funcAppSpec) == types.MethodType or\
       type(funcAppSpec) == types.BuiltinFunctionType or\
       type(funcAppSpec) == types.BuiltinMethodType:
      return(funcAppSpec)
    if type(funcAppSpec) == types.StringType:
      funcAppSpec = string.strip(funcAppSpec)
      if (funcAppSpec[0] <> '[') or (funcAppSpec[-1] <> ']'):
	self.handleError("Reference particle '" +\
			 funcAppSpec +\
			 "' looks like a function application spec but is not.",
			 BadReferenceParticle, funcAppSpec)
      # Turn the string '[<func>, <str>]' into an array with the actual func:
      arr = string.splitfields(funcAppSpec, ",")
      arr[0] = arr[0][1:]
      # Get the func name w/o the leading '[' and the fld name w/o the trailing ']':
      arr[1] = string.strip(arr[1][:-1])
      arr = [eval(arr[0]), arr[1]]
      if ((type(arr[0]) == types.FunctionType) or \
	  (type(arr[0]) == types.MethodType) or \
	  (type(arr[0]) == types.BuiltinFunctionType) or \
	  (type(arr[0]) == types.BuiltinMethodType)) and \
	  (type(arr[1]) == types.StringType):
	return(arr)
      else:
	self.handleError("Reference particle '" +\
			 funcAppSpec +\
			 "'. Must be a function and a field name string.",
			 BadReferenceParticle, funcAppSpec)
    else:
      self.handleError("Reference particle '" +\
		       str(funcAppSpec) +\
		       "' looks like a function application spec but is not.",
		       BadReferenceParticle, funcAppSpec)
      

  # Given a bib entry and an array with a function and a field name, get the
  # respective field value from the entry and pass it to the function. Return
  # what that function returned:

  def applyFnToFldValue(self, entry, fnPlusFldName, particle):
    fn = fnPlusFldName[0]
    fldName = fnPlusFldName[1]
    if (type(fldName) == types.StringType) and\
       type(fn) in [types.FunctionType, types.MethodType,
		    types.BuiltinFunctionType, types.BuiltinMethodType]:
      return(fn(bibutils.getEntryFldSafely(entry, fldName), entry, particle))
    else:
      self.handleError("Reference particle '" +\
		       fnPlusFldName +\
		       "'. Must be a function and a field name string.",
		       BadReferenceParticle, fnPlusFldName)

  # Returns 1 if self.lastSeparator is something recognized as requiring
  # capitalization. self.lastSeparator is an array of TaggedStr. In the vast
  # majority of the cases, this will have a single element which is a comma
  # or period, etc.:

  def shouldUpperCase(self):
    if (not self.lastSeparator):
      return(0)
    sepTstr = self.lastSeparator[-1].strip()
    if not sepTstr:
      return(0)
    sepChr = sepTstr[-1]
    return(sepChr in self.separatorsForCapitalization)

  def pageOrPages(self, pageInfo, entry, particle):
    if not string.strip(pageInfo):
      return([])
    if type(particle.fieldTag) == types.ListType:
      tags = particle.fieldTag + bibutils.getLinkTags(entry, 'pages')
    else:
      tags = [particle.fieldTag] + bibutils.getLinkTags(entry, 'pages')
    # If there are only digits in the page info, then it's only one
    # page that's being referenced. Else we assume the non-digit is
    # some form of range (like a dash, or an endash or an emdash):
    if bibutils.digitsOnly(pageInfo):
      if self.shouldUpperCase():
        return([TaggedStr(tags, 'Page ' + pageInfo)])
      else:
        return([TaggedStr(tags, 'p. ' + pageInfo)])
    else:
      if self.shouldUpperCase():
        return([TaggedStr(tags, 'Pages ' + pageInfo)])
      else:
        return([TaggedStr(tags, 'pp. ' + pageInfo)])

  def handleChapter(self, chapterInfo, entry, particle):
    if not string.strip(chapterInfo):
      return([])
    if type(particle.fieldTag) == types.ListType:
      tags = particle.fieldTag + bibutils.getLinkTags(entry, 'chapter')
    else:
      tags = [particle.fieldTag] + bibutils.getLinkTags(entry, 'chapter')
    if self.shouldUpperCase():
      res = 'Chapter '
    else:
      res = 'chapter '
    if bibutils.digitsOnly(chapterInfo):
      # Chapter is only numeric:
      res = res + chapterInfo
    else:
      res = res + '"' + chapterInfo + '"'
    return([TaggedStr(tags, res)])

  def handleNumber(self, numberInfo, entry, particle):
    if not string.strip(numberInfo):
      return([])
    if type(particle.fieldTag) == types.ListType:
      tags = particle.fieldTag + bibutils.getLinkTags(entry, 'number')
    else:
      tags = [particle.fieldTag] + bibutils.getLinkTags(entry, 'number')
    if self.shouldUpperCase():
      return([TaggedStr(tags, 'Number ' + numberInfo)])
    else:
      return([TaggedStr(tags, 'number ' + numberInfo)])

  def handleVolume(self, volumeInfo, entry, particle):
    if not string.strip(volumeInfo):
      return([])
    if type(particle.fieldTag) == types.ListType:
      tags = particle.fieldTag + bibutils.getLinkTags(entry, 'volume')
    else:
      tags = [particle.fieldTag] + bibutils.getLinkTags(entry, 'volume')
    if self.shouldUpperCase():
      return([TaggedStr(tags, 'Vol. ' + volumeInfo)])
    else:
      return([TaggedStr(tags, 'vol. ' + volumeInfo)])

  # Given an array of tagged strings, return a new array of tagged strings where
  # all adjacent strings with the same tag are combined. Empty strings are deleted
  # no matter what the tag:

  def compactTaggedStrings(self, originalTstrArray):
    if not originalTstrArray:
      return([])
    # Make copies of the tagged strings, because we'll do modifications that would
    # otherwise affect what was passed in:
    tstrArray = []
    for tstr in originalTstrArray:
      if tstr.str:
	tstrArray.append(copy.copy(tstr))
    res = [tstrArray[0]]
    for tstrIndex in range(len(tstrArray)-1):
      if tstrArray[1+tstrIndex].str:
        if res[-1].tag == tstrArray[1+tstrIndex].tag:
 	  res[-1].str = res[-1].str + tstrArray[1+tstrIndex].str
        else:
 	  res.append(tstrArray[tstrIndex+1])
    if res[0].str:
      return(res)
    else:
      return([])

  def classExists(self, className):
    try:
      if type(eval(className)) == types.ClassType:
	return(1)
      else:
	return(0)
    except:
      return(0)

######################################
#
# Generic Bibliography Style
#
######################################

# This pretty much does bibliographies according to Marie-Clair Van Leunen:
# "Handbook for Scholars", 1992.

class GenericStyle(ReferenceGenerator):
  styleName = 'GenericStyle'

  def __init__(self):
    ReferenceGenerator.__init__(self)

# These classes must be called <field-type>GenericStyle:

class articleGenericStyle(GenericStyle):
  
  def __init__(self):
    GenericStyle.__init__(self)

    self.journalParticle = RefParticle(
      fieldTag      = italic,
      fieldName     = 'journal',
      existenceReq  = mandatory)
				  
    self.volParticle = RefParticle(
      separator     = ', ',
      fieldName     = 'volume')
  
    self.numParticle = RefParticle(
      leader        = '(',
      fieldName     = 'number',
      trailer       = ')')

    self.pageParticle = RefParticle(
      leader        = ':',
      fieldName     = 'pages')

    self.constructPubInfoSpec(spec = [self.journalParticle,
		                      self.volParticle,
		                      self.numParticle,
		                      self.pageParticle,
		                      self.dateParticle])

class bookGenericStyle(GenericStyle):
  def __init__(self):
    GenericStyle.__init__(self)

    # Override the parent's title Particle to have tag italic:
    self.titleParticle  = RefParticle(
      fieldTag      = italic,
      fieldName     = 'booktitle _|_ title',
      trailer       = '. ',
      existenceReq  = mandatory)
  
    # Since we overrode the title particle, must call
    # the specification constructors for title spec again:
    self.constructTitleSpec()

class bookletGenericStyle(GenericStyle):

  def __init__(self):
    GenericStyle.__init__(self)

    # Take out requirement for author:
    self.authorParticle = RefParticle(
      fieldName     = '[self.fixAuthors, author] _|_ editor',
      trailer       = '. ')

    self.constructAuthorSpec()
  
class inbookGenericStyle(GenericStyle):
  booktitleParticle = 0

# This class is complicated. Must handle two different cases: referencing a piece
# within a collection, or the whole collection:
#
# A. Smith, editor. Collected Works of Byron. Addison-Wesley, 1992.
#   vs.
# K. Brown. My Life With Lucy. In A. Smith, editor, Collected Works of Byron,
#     pp 2-60. Addison-Wesley, 1978

  def __init__(self):
    GenericStyle.__init__(self)
  
    self.authorParticle = RefParticle(
      fieldName     = '[self.fixAuthors, author] _|_ [self.handleEditorNoAuthor, editor]',
      trailer       = '. ',
      existenceReq  = mandatory)

    self.titleParticle  = RefParticle(
      fieldName     = 'title _|_ [self.handleBookTitleOnly, booktitle]',
      trailer       = '. ',
      existenceReq  = mandatory)

    # The function rets an editor only if this refs a work within a collection.
    # In that case, the booktitle is added. This will produce: 
    #          'In Smith, editor, <booktitle>'

    self.booktitleParticle  = RefParticle(
      separator     = ', ',
      fieldTag      = italic,
      fieldName     = '[self.maybeHandleWorkWithinCollection, editor] _&&_ booktitle',
      conjunctSeparator = ' ')

    self.seriesParticle = RefParticle(
      separator     = '. ',
      fieldName     = 'series _&_ [self.handleVolume, volume]',
      conjunctSeparator = ', ',
      trailer       = '. ')

    self.editionParticle = RefParticle(
      separator     = '. ',
      fieldName     = 'edition')

    self.constructAuthorSpec()
    self.constructTitleSpec()
    self.constructPubInfoSpec()

  # Override ReferenceGenerator's constructPubInfoSpec method so that
  # classes inheriting from this generic inbook call this one:

  def constructPubInfoSpec(self, spec=[]):
    if spec:
      self.pubInfoSpec = spec
    else:
      self.pubInfoSpec = [self.booktitleParticle,
			  self.chapterParticle,
			  self.pageParticle,
			  self.seriesParticle,
			  self.publisherParticle,
			  self.editionParticle,
			  self.addressParticle,
			  self.dateParticle]

  # Called automatically with an editor field value. Produces
  # '<editor>, editor(s)'. Return an array of TaggedStr:

  def handleEditorNoAuthor(self, rawEditor, entry, particle):
    if not string.strip(rawEditor):
      return([])
    # fixAuthors returns an array of TaggedStr:
    editors = self.fixAuthors(rawEditor, entry, particle)
    linkTags = bibutils.getLinkTags(entry, 'editor')
    if linkTags:
      for editor in editors:
	editor.addTag(linkTags)
    if find(editors[0], 'and') > -1:
      return([editors[0] + ", editors"])
    else:
      return([editors[0] + ", editor"])
  
  # Called when inbook does not have a title for a piece of work within
  # the book, but only references the whole collection:

  def handleBookTitleOnly(self, booktitle, entry, particle):
    if not string.strip(booktitle):
      return([])
    tags = [italic] + bibutils.getLinkTags(entry, 'number')
    return([TaggedStr(tags, booktitle)])

  # Called to produce: 'In <editor>, editor(s), <booktitle> piece iff
  # there is a <title> value, indicating that this inbook ref points
  # to a piece within a collection:

  def maybeHandleWorkWithinCollection(self, editor, entry, particle):
    if bibutils.getEntryFldSafely(entry, 'title'):
      if self.shouldUpperCase():
        res = TaggedStr(plain, 'In ')
      else:
	res = TaggedStr(plain, 'in ')
      if string.strip(editor):
        # Get <editor>(s):
        res = res + self.handleEditorNoAuthor(editor, entry, particle)[0] + ', '
      return([res])
    else:
      return([])

class incollectionGenericStyle(inbookGenericStyle):
  pass

class inproceedingsGenericStyle(inbookGenericStyle):
  pass

class conferenceGenericStyle(inproceedingsGenericStyle):
  pass

class mastersthesisGenericStyle(bookGenericStyle):

  def __init__(self):
    bookGenericStyle.__init__(self)

    # Add advisory and the new leader showing it as a Master's. Thesis:
    self.publisherParticle  = RefParticle(
      separator     = '. ',
      leader        = "Master's thesis, ",
      fieldName     = 'publisher _|_ organization _|_ school _|_ institution',
      existenceReq  = advised)

    self.constructPubInfoSpec()

class miscGenericStyle(GenericStyle):
  def __init__(self):
    GenericStyle.__init__(self)

    # Take out requirement for author and title:
    self.authorParticle = RefParticle(
      fieldName     = '[self.fixAuthors, author] _|_ editor',
      trailer       = '. ')

    self.titleParticle  = RefParticle(
      fieldName     = 'title',
      trailer       = '. ')

    self.constructAuthorSpec()
    self.constructTitleSpec()

class phdthesisGenericStyle(bookGenericStyle):

  def __init__(self):
    bookGenericStyle.__init__(self)

    # Add advisory and the new leader showing it as a PhD. Thesis:
    self.publisherParticle  = RefParticle(
      separator     = '. ',
      leader        = 'PhD thesis, ',
      fieldName     = 'publisher _|_ organization _|_ school _|_ institution',
      existenceReq  = advised)

    self.constructPubInfoSpec()

class techreportGenericStyle(bookGenericStyle):

  def __init__(self):
    bookGenericStyle.__init__(self)

    # Add existence advisory to techreport number:
    self.numParticle = RefParticle(
      separator     = ', ',
      fieldName     = '[self.handleNumber, number]',
      existenceReq  = advised)

    # Add existence advisory to publisher:
    self.publisherParticle  = RefParticle(
      separator     = '. ',
      fieldName     = 'publisher _|_ organization _|_ school _|_ institution',
      existenceReq  = advised)

    self.constructPubInfoSpec()

  # Override ReferenceGenerator's constructPubInfoSpec method so that
  # classes inheriting from this class call this one:

  def constructPubInfoSpec(self, spec=[]):
    if spec:
      self.pubInfoSpec = spec
    else:
      self.pubInfoSpec = [self.seriesParticle,
			  self.numParticle,
			  self.pageParticle,
                          self.publisherParticle,
			  self.addressParticle,
			  self.dateParticle]

class manualGenericStyle(techreportGenericStyle):

  def __init__(self):
    techreportGenericStyle.__init__(self)

    # Override the parent's author Particle to allow missing author:
    self.authorParticle = RefParticle(
      fieldName     = '[self.fixAuthors, author] _|_ editor',
      trailer       = '. ')

    # Take away existence advisory from techreport number:
    self.numParticle = RefParticle(
      separator     = ', ',
      fieldName     = '[self.handleNumber, number]')

    # Since we overrode the author and a pubInfo particle, must call
    # the specification constructors for title and pubInfo specs again:
    self.constructAuthorSpec()
    self.constructPubInfoSpec()

class unpublishedGenericStyle(GenericStyle):

  def __init__(self):
    GenericStyle.__init__(self)

    # Add existence advisory to note:
    self.notesParticle  = RefParticle(
      separator     = '. ',
      fieldName     = 'howpublished _&_ note',
      existenceReq  = advised)

    self.constructNotesSpec()

######################################
#
# IEEE Bibliography Style
#
######################################
    
class IEEEStyle(ReferenceGenerator):
  styleName = 'IEEEStyle'

  def __init__(self):
    ReferenceGenerator.__init__(self)

    self.addressParticle  = RefParticle(
      separator     = '. ',
      fieldName     = 'address',
      trailer       = ': ')              # For 'New York: Wesley'

# These classes must be called <field-type>IEEEStyle:

class articleIEEEStyle(IEEEStyle):
  
  def __init__(self):
    IEEEStyle.__init__(self)

    # Override the parent's author Particle to have a comma separating it from the title:

    self.authorParticle  = RefParticle(
      fieldName     = 'author',
      trailer       = ', ',
      existenceReq  = mandatory)

    # Override the parent's title Particle to have tag plain and double quotes:
    self.titleParticle  = RefParticle(
      leader        = '"',
      fieldName     = 'title',
      trailer       = ',"',
      existenceReq  = mandatory)
  
    self.journalParticle = RefParticle(
      fieldTag      = italic,
      fieldName     = 'journal',
      existenceReq  = mandatory)
				  
    self.volParticle = RefParticle(
      separator     = ', ',
      fieldName     = '[self.handleVolume, volume]')

    self.constructAuthorSpec()
    self.constructTitleSpec()
    self.constructPubInfoSpec([self.journalParticle,
		               self.volParticle,
		               self.numParticle,
		               self.pageParticle,
		               self.dateParticle])

class bookIEEEStyle(bookGenericStyle, IEEEStyle):

  def __init__(self):
    bookGenericStyle.__init__(self)
    IEEEStyle.__init__(self)

    self.titleParticle  = RefParticle(
      fieldTag      = italic,
      fieldName     = 'booktitle _|_ title',
      trailer       = ', ',
      existenceReq  = mandatory)

    self.constructTitleSpec()
    # This calls this class' pub spec constructor:
    self.constructPubInfoSpec()

  # Since we inherit from this class, override the pub info constructor:

  def constructPubInfoSpec(self, spec=[]):
    if spec:
      self.pubInfoSpec = spec
    else:
      self.pubInfoSpec = [self.seriesParticle,
			  self.editionParticle,
			  self.addressParticle,
			  self.publisherParticle,
			  self.dateParticle,
			  self.pageParticle]


class bookletIEEEStyle(IEEEStyle):

  def __init__(self):
    IEEEStyle.__init__(self)

    # Take out requirement for author:
    self.authorParticle = RefParticle(
      fieldName     = '[self.fixAuthors, author] _|_ editor',
      trailer       = '. ')

    self.constructAuthorSpec()

class inbookIEEEStyle(inbookGenericStyle, IEEEStyle):

  def __init__(self):
    inbookGenericStyle.__init__(self)
    IEEEStyle.__init__(self)
    
    self.authorParticle = RefParticle(
      fieldName     = '[self.fixAuthors, author] _|_ [self.handleEditorNoAuthor, editor]',
      trailer       = ', ',
      existenceReq  = mandatory)
    
    self.titleParticle  = RefParticle(
      leader        = '"',
      fieldName     = 'title _|_ [self.handleBookTitleOnly, booktitle]',
      trailer       = '," ',
      existenceReq  = mandatory)
  
    self.booktitleParticle  = RefParticle(
      separator     = ', ',
      fieldName     = '[self.maybeHandleWorkWithinCollection, editor] _&&_ \
                       [self.handleEditorNoAuthor, editor]',
      conjunctSeparator = ', ',
      trailer       = '. ')

    self.dateParticle = RefParticle(
      separator     = ', ',      
      fieldName     = 'month _&_ year',
      conjunctSeparator = ', ')


    self.constructAuthorSpec()
    self.constructTitleSpec()
    self.constructPubInfoSpec([self.booktitleParticle,
			       self.chapterParticle,
			       self.seriesParticle,
			       self.addressParticle,
			       self.publisherParticle,
			       self.dateParticle,
			       self.pageParticle])
			       
  # If there's a title field AND a booktitle field, assemble "In <booktitle>":

  def maybeHandleWorkWithinCollection(self, editor, entry, particle):
    res = []
    tags = [italic] + bibutils.getLinkTags(entry, 'booktitle')
    if bibutils.getEntryFldSafely(entry, 'title'):
      booktitle = bibutils.getEntryFldSafely(entry, 'booktitle')
      if not booktitle:
	return(res)
      if self.shouldUpperCase():
        res.append(TaggedStr(plain, 'In '))
      else:
	res.append(TaggedStr(plain, 'in '))
      res.append(TaggedStr(tags, booktitle))
      return(res)
    else:
      return([])

class incollectionIEEEStyle(inbookIEEEStyle):
  pass

class inproceedingsIEEEStyle(inbookIEEEStyle):
  pass

class conferenceIEEEStyle(inproceedingsIEEEStyle):
  pass

class mastersthesisIEEEStyle(bookIEEEStyle):

  def __init__(self):
    bookIEEEStyle.__init__(self)

    # Add advisory and the new leader showing it as a Master's. Thesis:
    self.publisherParticle  = RefParticle(
      separator     = '. ',
      leader        = "Master's thesis, ",
      fieldName     = 'publisher _|_ organization _|_ school _|_ institution',
      existenceReq  = advised)

    self.constructPubInfoSpec()

class miscIEEEStyle(IEEEStyle):

  def __init__(self):
    IEEEStyle.__init__(self)

    # Take out requirement for author and title:
    self.authorParticle = RefParticle(
      fieldName     = '[self.fixAuthors, author] _|_ editor',
      trailer       = '. ')

    self.titleParticle  = RefParticle(
      fieldName     = 'title',
      trailer       = '. ')

    self.constructAuthorSpec()
    self.constructTitleSpec()


class phdthesisIEEEStyle(bookIEEEStyle):

  def __init__(self):
    bookIEEEStyle.__init__(self)

    # Add advisory and the new leader showing it as a PhD. Thesis:
    self.publisherParticle  = RefParticle(
      separator     = '. ',
      leader        = 'PhD thesis, ',
      fieldName     = 'publisher _|_ organization _|_ school _|_ institution',
      existenceReq  = advised)

    self.constructPubInfoSpec()

class techreportIEEEStyle(bookIEEEStyle):

  def __init__(self):
    bookIEEEStyle.__init__(self)

    # Add existence advisory to techreport number:
    self.numParticle = RefParticle(
      separator     = ', ',
      fieldName     = '[self.handleNumber, number]',
      existenceReq  = advised)

    # Add existence advisory to publisher:
    self.publisherParticle  = RefParticle(
      separator     = '. ',
      fieldName     = 'publisher _|_ organization _|_ school _|_ institution',
      existenceReq  = advised)

    self.constructPubInfoSpec()
    self.constructPubInfoSpec()

  # Since we inherit from this class, override the pub info constructor:

  def constructPubInfoSpec(self, spec=[]):
    if spec:
      self.pubInfoSpec = spec
    else:
      self.pubInfoSpec = [self.seriesParticle,
			  self.numParticle,
			  self.addressParticle,
                          self.publisherParticle,
			  self.dateParticle,
			  self.pageParticle]

class manualIEEEStyle(techreportIEEEStyle):

  def __init__(self):
    techreportIEEEStyle.__init__(self)

    # Override the parent's author Particle to allow missing author:
    self.authorParticle = RefParticle(
      fieldName     = '[self.fixAuthors, author] _|_ editor',
      trailer       = '. ')

    # Take away existence advisory from techreport number:
    self.numParticle = RefParticle(
      separator     = ', ',
      fieldName     = '[self.handleNumber, number]')

    # Since we overrode the author and a pubInfo particle, must call
    # the specification constructors for title and pubInfo specs again:

    self.constructAuthorSpec()
    self.constructPubInfoSpec()

class unpublishedIEEEStyle(IEEEStyle):

  def __init__(self):
    IEEEStyle.__init__(self)

    # Add existence advisory to note:
    self.notesParticle  = RefParticle(
      separator     = '. ',
      fieldName     = 'howpublished _&_ note',
      existenceReq  = advised)

    self.constructNotesSpec()

######################################
#
# FullAuthor Bibliography Style
#
######################################
    
# This class does everything like GenericStyle, but the date is made
# part of the author, and it first looks in field yearKey for
# the date spec. This is for citations such as:
#
#   As we see in (Miller, Smith and Mayer, 1985), ....
#   and also in (Miller, Smith and Mayer, 1985a).
#
# with reference list entries like:
#
#   Frank Miller, Ian Smith and Gloria Mayer (1985). The First Amendment...
#   Frank Miller, Ian Smith and Gloria Mayer (1985a). The Second Amendment...
#
# We expect yearKey to have the year, possibly with a disambiguating letter
# attached as in the second entry. If yearKey isn't there, it uses
# the regular year.
# After it uses the yearKey, it clears it out to avoid accidental reuse
# later when the entry is used in a different context.

class FullAuthorStyle(GenericStyle):
  styleName = 'FullAuthorStyle'

  def __init__(self):
    # DONT call GenericStyle.__init__(self) here! All the subclasses
    # call carefully selected subclasses of GenericStyle and THEN
    # call this init method. If we call the generic init here, we
    # destroy any overriding we did earlier.

    self.authorParticle = RefParticle(
      fieldName     = '[self.handleFullAuthor, author] _|_ [self.handleFullAuthor, editor]',
      trailer       = '. ',
      existenceReq  = mandatory)

    self.constructAuthorSpec()

    self.constructPubInfoSpec([self.pageParticle,
	 		       self.seriesParticle,
			       self.numParticle,
			       self.publisherParticle,
			       self.editionParticle,
		               self.addressParticle])

  def fixYearKey(self, yearKey, entry, particle):
    if not yearKey:
      return([])
    res = TaggedStr(plain, entry.yearKey)
    if hasattr(entry, 'yearKey'):
      delattr(entry, 'yearKey')
    return([res])

  def handleFullAuthor(self, author, entry, particle):
    if not author:
      return([])
    fixedAuthors = self.fixAuthors(author, entry, particle)
    year = bibutils.getEntryFldSafely(entry, 'yearKey')
    if year:
      delattr(entry, 'yearKey')
    else:
      year = bibutils.getEntryFldSafely(entry, 'year')
    if year:
      return([fixedAuthors[0] + ' (' + year + ')'])
    else:
      return([fixedAuthors[0]])

  def handleFullTitle(self, title, entry, particle):
    if not title:
      return([])
    if bibutils.getEntryFldSafely(entry, 'author') or\
       bibutils.getEntryFldSafely(entry, 'editor'):
      year = ''
    elif bibutils.getEntryFldSafely(entry, 'yearKey'):
      year = bibutils.getEntryFldSafely(entry, 'yearKey')
      delattr(entry, 'yearKey')
    else:
      year = bibutils.getEntryFldSafely(entry, 'year')
    if year:
      return([TaggedStr(plain, title + ' (' + year + ')')])
    else:
      return([TaggedStr(plain, title)])

# These classes must be called <field-type>IEEEStyle:

class articleFullAuthorStyle(articleGenericStyle, FullAuthorStyle):
  def __init__(self):
    articleGenericStyle.__init__(self)
    FullAuthorStyle.__init__(self)

    self.constructPubInfoSpec(spec = [self.journalParticle,
		                      self.volParticle,
		                      self.numParticle,
		                      self.pageParticle])

class bookFullAuthorStyle(bookGenericStyle, FullAuthorStyle):
  def __init__(self):
    bookGenericStyle.__init__(self)
    FullAuthorStyle.__init__(self)

class bookletFullAuthorStyle(bookletGenericStyle, FullAuthorStyle):
  def __init__(self):
    bookletGenericStyle.__init__(self)
    FullAuthorStyle.__init__(self)

    # This is allowed to not have an author. So need to do something
    # special with the year that doubles as a key:
    
    self.authorParticle = RefParticle(
      fieldName     = '[self.handleFullAuthor, author] _|_ [self.handleFullAuthor, editor]',
      trailer       = '. ')

    self.titleParticle  = RefParticle(
      fieldName     = '[self.handleFullTitle, title]',
      trailer       = '. ')

    self.constructAuthorSpec()
    self.constructTitleSpec()


class inbookFullAuthorStyle(inbookGenericStyle, FullAuthorStyle):
  def __init__(self):
    inbookGenericStyle.__init__(self)
    FullAuthorStyle.__init__(self)

    self.constructPubInfoSpec()

  # Override GenericStyle's constructPubInfoSpec method so that
  # classes inheriting from this inbook call this one:

  def constructPubInfoSpec(self, spec=[]):
    if spec:
      self.pubInfoSpec = spec
    else:
      self.pubInfoSpec = [self.booktitleParticle,
			  self.chapterParticle,
			  self.pageParticle,
			  self.seriesParticle,
			  self.publisherParticle,
			  self.editionParticle,
			  self.addressParticle]
    
class incollectionFullAuthorStyle(inbookFullAuthorStyle):
  pass

class inproceedingsFullAuthorStyle(inbookFullAuthorStyle):
  pass

class conferenceFullAuthorStyle(inproceedingsFullAuthorStyle):
  pass

class mastersthesisFullAuthorStyle(mastersthesisGenericStyle, FullAuthorStyle):
  def __init__(self):
    mastersthesisGenericStyle.__init__(self)
    FullAuthorStyle.__init__(self)

class miscFullAuthorStyle(miscGenericStyle, FullAuthorStyle):
  def __init__(self):
    miscGenericStyle.__init__(self)
    FullAuthorStyle.__init__(self)

    # This is allowed to not have an author. So need to do something
    # special with the year that doubles as a key:
    
    self.authorParticle = RefParticle(
      fieldName     = '[self.handleFullAuthor, author] _|_ [self.handleFullAuthor, editor]',
      trailer       = '. ')

    self.titleParticle  = RefParticle(
      fieldName     = '[self.handleFullTitle, title]',
      trailer       = '. ')

    self.constructAuthorSpec()
    self.constructTitleSpec()

class phdthesisFullAuthorStyle(phdthesisGenericStyle, FullAuthorStyle):
  def __init__(self):
    phdthesisGenericStyle.__init__(self)
    FullAuthorStyle.__init__(self)

class techreportFullAuthorStyle(techreportGenericStyle, FullAuthorStyle):
  def __init__(self):
    techreportGenericStyle.__init__(self)
    FullAuthorStyle.__init__(self)

    self.constructPubInfoSpec()
    
  # Override GenericStyle's constructPubInfoSpec method so that
  # classes inheriting from this class call this one:

  def constructPubInfoSpec(self, spec=[]):
    if spec:
      self.pubInfoSpec = spec
    else:
      self.pubInfoSpec = [self.seriesParticle,
			  self.numParticle,
			  self.pageParticle,
                          self.publisherParticle,
			  self.addressParticle]

class manualFullAuthorStyle(techreportFullAuthorStyle):
  def __init__(self):
    techreportFullAuthorStyle.__init__(self)

    # This is allowed to not have an author. So need to do something
    # special with the year that doubles as a key:
    
    self.authorParticle = RefParticle(
      fieldName     = '[self.handleFullAuthor, author] _|_ [self.handleFullAuthor, editor]',
      trailer       = '. ')

    self.titleParticle  = RefParticle(
      fieldName     = '[self.handleFullTitle, title]',
      trailer       = '. ')

    # Take away existence advisory from techreport number:
    self.numParticle = RefParticle(
      separator     = ', ',
      fieldName     = '[self.handleNumber, number]')

    self.constructAuthorSpec()
    self.constructTitleSpec()
    self.constructPubInfoSpec()

class unpublishedFullAuthorStyle(unpublishedGenericStyle, FullAuthorStyle):
  def __init__(self):
    unpublishedGenericStyle.__init__(self)
    FullAuthorStyle.__init__(self)


######################################
#
# Service Classes
#
######################################

# TaggedStr objects have an array of tags plus a regular string. You can set
# tags individually, or you can set a whole array of them:
#
#   foo = TaggedStr(italic, 'Blue Ship')
#   foo = TaggedStr([bold, italic], 'Blue Ship')
#
# both work.
#
# You can treat TaggedStr's pretty much like regular strings: you add them to
# get a new TaggedStr that's the concatenation. This fails if the two strings
# have different tags. You can get a slice: foo[1:3] will get a new TaggedStr
# which is the slice of foo. You can compare two TaggedStr and get the expected
# result (tags are ignored for comparison ops). You can do len(foo) and get foo's
# string length. You can do something like:
#
#   if foo:
#     print 'Foo has something in it'
#
# If foo has a zero length string part, the IF test will fail properly. There are
# also many useful methods, like lower, upper, find, etc.

class TaggedStr:
  tag = [plain]
  str = ""

  def __init__(self, theTag=plain, theStr=""):
    # We need to set the tag once to get it into the dictionnary.
    # Otherwise addTag, and therefore all tag setting will fail:
    self.tag = theTag
    self.str = theStr

  def find(self, sub, start=0):
    return(string.find(self.str, sub, start))

  def rfind(self, sub, start=0):
    return(string.rfind(self.str, sub, start))

  def index(self, sub, start=0):
    return(string.index(self.str, sub, start))

  def rindex(self, sub, start=0):
    return(string.rindex(self.str, sub, start))

  def count(self, sub, start=0):
    return(string.count(self.str, sub, start))

  def lower(self):
    return(TaggedStr(self.tag, string.lower(self.str)))

  def strip(self):
    return(TaggedStr(self.tag, string.strip(self.str)))

  def swapcase(self):
    return(TaggedStr(self.tag, string.swapcase(self.str)))

  def upper(self):
    return(TaggedStr(self.tag, string.upper(self.str)))

  def ljust(self, width):
    return(TaggedStr(self.tag, ljust(self.str, width)))
    
  def rjust(self, width):
    return(TaggedStr(self.tag, rjust(self.str, width)))

  def center(self, width):
    return(TaggedStr(self.tag, center(self.str, width)))

  def zfill(self, width):
    return(TaggedStr(self.tag, zfill(self.str, width)))

  def __setattr__(self, name, value):
    # Allow putting tags in as either strings or array of strings.
    # Convert all to an array, though:
    if (name == 'tag'):
      self.__dict__['tag'] = []
      # Go through addTag to have tag error checking all in one place:
      self.addTag(value)
    else:
      self.__dict__[name] = value
    
  # Given a single tag or a list of them, check tag sanity and add it/them
  # to the list of tags on the string. NOTE: This is called from
  # __setattr__, so you can't set the tag attribute through normal means.
  # You'd recurse infinitely:

  def addTag(self, tags):
    if type(tags) <> types.ListType:
      tags = [tags]
    # Got a list of tags to append. Remove duplicates:
    for newTag in tags:
      if newTag not in self.tag:
	if (type(newTag) == types.StringType) or\
	   (type(newTag) == types.InstanceType):
	  self.__dict__['tag'].append(newTag)

  def delTag(self, tags):
    if type(tags) == types.ListType:
      for oldTag in tags:
        if oldTag in self.tag:
	  self.__dict__['tag'].remove(oldTag)
    else:
      self.__dict__['tag'].remove(tags)
    if not self.tag:
      self.addTag(plain)

  def blank(self):
    string.strip(self.str)

  def __nonzero__(self):
    return(len(self.str))

  def __str__(self):
    if len(self.str) == 0:
      return("")
    tags = '['
    for aTag in self.tag:
      # If it's a link:
      if type(aTag) == types.InstanceType:
	tags = tags + str(aTag)
      if aTag == italic:
	tags = tags + '\\i'
      elif aTag == bold:
	tags = tags + '\\b'
      elif aTag == plain:
	continue
      else:
	raise BadTagValue, aTag
    tags = tags + ']'
    if len(tags) > 2:
      return tags + self.str + '[\\]'
    else:
      return self.str

  def __repr__(self):
    res = 'TaggedStr(['
    for aTag in self.tag:
      pos = len(res)
      # If tag is a link:
      if type(aTag) == types.InstanceType:
	aTag = str(aTag)
      res = res + aTag
      res = res[:pos] + res[pos:pos+len(aTag)] + ', '
    res = res[:-2] + '], "' + self.str + '")'
    return(res)

  def __cmp__(self, other):
    if self.str < other.str:
      return(-1)
    if self.str == other.str:
      return(0)
    if self.str > other.str:
      return(1)
    
  def __getslice__(self, i, j):
    return(TaggedStr(self.tag, self.str[i:j]))
  
  def __getitem__(self, key):
    if type(key) <> types.IntType:
      raise TypeError, "sequence subscript not int"
    return(self.str[key])

  def __len__(self):
    return(len(self.str))

  # Takes either a TaggedStr or a regular string and returns
  # a new TaggedStr that's the concatenation. If tags are
  # not compatible: exception:
  
  def __add__(self, otherStr):
    if type(otherStr) == types.StringType:
      return(TaggedStr(self.tag, self.str + otherStr))
    else:
      return(self.tryMergeTags(otherStr))

  # Merge two tagged strings. We fail if the tags are different sets, except
  # if the only difference is that one of the tag sets includes a link. In
  # that case the returned tagged str will include the link. If both tags
  # contain links that are not all identical, we throw an exception:

  def tryMergeTags(self, otherStr):
    tmpSelfTags = []
    tmpOtherTags = []
    links = []
    for tag in self.tag:
      if type(tag) == types.InstanceType:
	links.append(tag)
      else:
	tmpSelfTags.append(tag)
    for tag in otherStr.tag:
      if type(tag) == types.InstanceType:
	links.append(tag)
      else:
	tmpOtherTags.append(tag)
    tmpSelfTags.sort()
    tmpOtherTags.sort()
    if tmpSelfTags <> tmpOtherTags:
      raise IncompatibleTags,\
	    str(self.tag) + self.str + ", " + str(otherStr.tag) + otherStr.str
    if len(links) > 1:
      # Check whether all links are the same:
      firstLinkStr = links[0].ascii();
      for link in links[1:]:
	if link.ascii() <> firstLinkStr:
          raise IncompatibleTags, "Merging strings with different links: " + \
	        str(self.tag) + self.str + ", " + str(otherStr.tag) + otherStr.str
      links = [links[0]]
    return(TaggedStr(tmpSelfTags + links, self.str + otherStr.str))

  def __setslice__(self, i, j, seq):
    raise TypeError, 'unassignable slice'

######################################
#
# Service Functions
#
######################################

def find(s, sub, start=0):
  if type(s) == types.StringType:
    return(string.find(s, sub, start))
  else:
    return(s.find(sub, start))
